"""Exceptions for errors."""

from __future__ import annotations

from ..exceptions import (  # noqa: F401
    CannotConnect,
    InvalidAuth,
    RequireValidation,
    YaleApiError,
    YaleXSError,
)
